"""
Benchmark result retrieval utilities.

Provides access to benchmarking metadata and environment variables,
aggregated as pandas DataFrames. Supports filtering by date, exclusions,
and time intervals.

Used for analysis, reporting, and visualization of benchmark data.
"""

from datetime import datetime
import pandas as pd
import yaml
import warnings

from opa_tools.config import BENCHMARK_DIR
from opa_tools.utils import process_dct, DATAFRAME_TYPES


def _select_directories_by_dates(benchmark_dir, dates=None):
    """
    Select directories matching specific date prefixes (if any).
    """
    directories = sorted(benchmark_dir.iterdir(), key=lambda path: path.name)
    if dates:
        directories = [
            d for d in directories if any(d.name.startswith(date) for date in dates)
        ]
    return directories


def _get_exclude_fun(exclude, directories):
    """
    Returns a function that filters out excluded benchmarks.
    Only works with single-date exclusions (string).
    """
    if exclude is None:
        return lambda _: False
    if len(directories) == 1:
        if isinstance(exclude, str):
            return lambda path: exclude == path.name
        else:
            raise NotImplementedError("Exclude must be a str for a single date.")
    raise NotImplementedError("Exclude works only with one date.")


def _get_interval_fun(interval):
    """
    Returns a filter function that selects benchmarks by datetime interval.
    """
    def is_in_interval(benchmark_dir):
        benchmark_date_file = benchmark_dir / "date.txt"
        if not benchmark_date_file.exists():
            warnings.warn(f"File {benchmark_dir / 'date.txt'} not found.")
            return False
        with open(benchmark_date_file, "r") as datefile:
            date = datetime.fromisoformat(datefile.read())
        return interval[0] <= date < interval[1]

    return is_in_interval if interval else lambda _: True


def get_benchmark_iterator(dates=None, exclude=None, interval=None):
    """
    Yields benchmark directories that pass all filters.
    """
    directories = _select_directories_by_dates(BENCHMARK_DIR, dates)
    exclude_fun = _get_exclude_fun(exclude, directories)
    interval_fun = _get_interval_fun(interval)

    for date_dir in directories:
        if (date_dir / ".SKIP").exists():
            continue
        for benchmark_dir in sorted(date_dir.iterdir()):
            if (
                (benchmark_dir / ".SKIP").exists() or
                exclude_fun(benchmark_dir) or
                not interval_fun(benchmark_dir)
            ):
                continue
            yield benchmark_dir


def get_environ_df(dates=None, exclude=None, interval=None):
    """
    Returns a DataFrame containing benchmark environment variables.
    """
    environ_df = pd.DataFrame()
    for benchmark_dir in get_benchmark_iterator(dates, exclude, interval):
        with (
            open(benchmark_dir / "date.txt", "r") as date_file,
            open(benchmark_dir / "environ.yaml", "r") as environ_file,
        ):
            try:
                environ_dict = yaml.safe_load(environ_file)
            except yaml.constructor.ConstructorError:
                continue
            dct = {
                "date": date_file.read(),
                "bname": benchmark_dir.name,
                **process_dct(environ_dict, valid_types=DATAFRAME_TYPES, list_wrap=True),
            }
            environ_df = pd.concat([environ_df, pd.DataFrame.from_dict(dct)], ignore_index=True)
    return environ_df


def get_request_df(dates=None, exclude=None, interval=None):
    """
    Returns a DataFrame containing benchmark request metadata.
    """
    request_df = pd.DataFrame()
    for benchmark_dir in get_benchmark_iterator(dates, exclude, interval):
        benchmark_dct = {
            "date": (benchmark_dir / "date.txt").read_text(),
            "bname": benchmark_dir.name,
        }
        for file in (benchmark_dir / "additional").iterdir():
            if "request" not in file.name:
                continue
            prefix = file.stem.removesuffix("request") + "_"
            with open(file, "r") as req_f:
                req_dct = yaml.safe_load(req_f)
                benchmark_dct.update(
                    process_dct(req_dct, prefix=prefix, valid_types=DATAFRAME_TYPES, list_wrap=True)
                )
        request_df = pd.concat([request_df, pd.DataFrame.from_dict(benchmark_dct)], ignore_index=True)
    return request_df
