from argparse import ArgumentParser
from pandas import DataFrame
from typing import Sequence
import sys
import warnings

from opa_tools.request import (
    create_template,
    generate_requests,
    get_request,
    get_request_df,
    get_request_identifiers,
)
from .utils import format_dataframe


def _get_parser():
    parser = ArgumentParser(
        prog="req",
        description=(
            "OPA Request CLI Utility\n\n"
            "This tool provides access to request-related data, allowing users to display, "
            "list, generate, copy, and create templates for requests.\n\n"
            "Identifiers may be:\n"
            "  - A month (YYMM), e.g. '2403'\n"
            "  - A request ID, e.g. 'abc123'\n"
            "  - A full reference, e.g. '2403/abc123'\n"
            "  - Wildcards like '*' may be supported for listing.\n"
        ),
        epilog="Run `req <command> -h` for command-specific options.",
    )
    subparsers = parser.add_subparsers(title="Available Commands", dest="service")

    # SHOW
    parser_show = subparsers.add_parser(
        "show",
        help="Display a table of requests for a given month.",
        description="Show a table of requests. Optionally filter columns and choose output format.",
    )
    _add_table_arguments(parser_show)

    # LIST
    parser_ls = subparsers.add_parser(
        "ls",
        help="List request identifiers.",
        description=(
            "List all request identifiers for a given month or pattern.\n\n"
            "Examples:\n"
            "  req ls 2403           # All requests in March 2024\n"
            "  req ls abc123         # Request 'abc123' (if globally unique)\n"
            "  req ls 2403/abc123    # Specific request in March 2024"
        )
    )
    parser_ls.add_argument(
        "identifier",
        nargs="?",
        default="*",
        help="Month of requests or request identifier (e.g., '2403', 'abc123', or '2403/abc123')",
    )
    parser_ls.add_argument(
        "-s", "--summarize",
        action="store_true",
        help="Collapse subrequests and show as a single request (e.g., 'abc123.*').",
    )

    # COLUMNS (Not implemented)
    parser_col = subparsers.add_parser(
        "col",
        help="(WIP) List available request columns.",
        description="List available columns in request data. [Not yet implemented]"
    )

    # GENERATE
    parser_gen = subparsers.add_parser(
        "gen",
        help="Generate request(s) for a given identifier.",
        description="Generate one or more requests using the given identifier (e.g., month or full ID)."
    )
    parser_gen.add_argument("identifier", type=str, action="store", help="Identifier to generate requests for.")
    parser_gen.add_argument("-f", "--force", action="store_true", help="Remove existing subrequest identifiers if they exist.")

    # COPY
    parser_cp = subparsers.add_parser(
        "cp",
        help="Copy a request from one identifier to another.",
        description="Copy an existing request to a new destination. Recursive copy not yet supported.",
    )
    parser_cp.add_argument(
        "-r", "--recursive",
        action="store_true",
        help="Recursively copy subrequests. [Not implemented]",
    )
    parser_cp.add_argument("src", type=str, help="Source request identifier.")
    parser_cp.add_argument("dst", type=str, help="Destination request identifier.")

    # TEMPLATE
    parser_template = subparsers.add_parser(
        "template",
        help="Create a request template from existing sources.",
        description="Generate a new request template using one or more source identifiers."
    )
    parser_template.add_argument("dst", type=str, help="Destination for the new template.")
    parser_template.add_argument(
        "--from", dest="src", type=str, nargs="+", required=True,
        help="Source identifiers to build the template from.",
    )

    return parser


def _add_table_arguments(parser):
    parser.add_argument(
        "-c", "--columns",
        action="store", nargs="*", default=[],
        help="Specific columns to display (e.g., 'id', 'bname', 'date')."
    )
    parser.add_argument(
        "month",
        action="store", nargs="?", default="*",
        help="Month to display requests for (e.g., '2403'). Use '*' for all."
    )
    parser.add_argument(
        "-f", "--format",
        action="store",
        type=str,
        choices=["default", "fwf"] + [
            method.lstrip("to_")
            for method in dir(DataFrame)
            if method.startswith("to_")
        ],
        default="default",
        help="Output format. Can be 'default', 'fwf', or any supported Pandas format (e.g., 'csv', 'json')."
    )
    parser.add_argument(
        "-t", "--tablefmt",
        action="store",
        type=str,
        metavar="TABULAR_TABLEFMT",
        default="plain",
        help="Tabular format style (e.g., 'plain', 'grid', 'fancy_grid'). Only used with some formats.",
    )


def _validate_columns(requested, existing):
    def force_column(lst, col):
        if col in existing:
            if col in lst:
                return
            lst.insert(0, col)
    columns = requested.copy()
    force_column(columns, "date")
    force_column(columns, "bname")
    wrong_columns = list(filter(lambda x: x not in existing, columns))
    if wrong_columns:
        warnings.warn(
            f"Column(s) {', '.join(wrong_columns)} skipped."
        )
        columns = [column for column in columns if column not in wrong_columns]
    return columns

def _normalize_identifiers(identifiers: Sequence[str]):
    # TODO: Add validation and glob expansion
    return identifiers


def _table_service(args):
    df = get_request_df(month=args.month)
    if args.columns:
        columns = _validate_columns(args.columns, df.columns)
        df = df[columns]
    print(format_dataframe(df, args.format))


def _list_service(args):
    identifier = args.identifier
    summarize = args.summarize
    identifiers = get_request_identifiers(month=identifier, subrequests=True)
    if not summarize:
        print("\n".join(identifiers))
        return
    identifier_ns = get_request_identifiers(month=identifier, subrequests=False)
    for i, request in enumerate(identifier_ns):
        if request not in identifiers:
            identifiers = list(filter(
                lambda x: not str.startswith(x, request),
                identifiers
            ))
            identifiers.insert(i, f"{request}.*")
    print("\n".join(identifiers))


def _column_service(args):
    raise NotImplementedError("The 'col' service is not yet implemented.")
    print("\n".join(sorted(df.columns)))


def _generate_service(args):
    generate_requests(args.identifier, args.force)


def _copy_service(args):
    if args.recursive:
        raise NotImplementedError("Recursive copy is not yet implemented.")
    request = get_request(args.src)
    raise NotImplementedError("Copy functionality is a stub.")


def _template_service(args):
    create_template(src=args.src, dst=args.dst)


_SERVICES = {
    "show": _table_service,
    "ls": _list_service,
    "col": _column_service,
    "gen": _generate_service,
    "cp": _copy_service,
    "template": _template_service,
}


def req(argv=None):
    if argv is None:
        argv = sys.argv[1:]

    parser = _get_parser()
    args = parser.parse_args(argv)
    try:
        service = _SERVICES[args.service]
    except KeyError:
        parser.print_help()
        sys.exit(1)
    service(args)
