from argparse import ArgumentParser, RawDescriptionHelpFormatter
from concurrent.futures import as_completed, ProcessPoolExecutor
from datetime import datetime
from pathlib import Path
import pickle
import time
from typing import TYPE_CHECKING
import typing as T
import os
import sys
import yaml
from pprint import pprint

import numpy as np
from numpy.typing import NDArray

from crick import TDigest
try:
    from crick import TDigestArray
except ImportError:
    TDigestArray = None

# Check if we are under an MPI environment
MPI_LAUNCHED = any(var in os.environ for var in [
    "OMPI_COMM_WORLD_SIZE",  # OpenMPI
    "PMI_SIZE",              # MPICH / IntelMPI
    "MPI_LOCALNRANKS",       # Some others
])

if MPI_LAUNCHED:
    try:
        from mpi4py import MPI
        MPI_COMM = MPI.COMM_WORLD
        MPI_RANK = MPI_COMM.Get_rank()
        MPI_SIZE = MPI_COMM.Get_size()
    except ImportError:
        MPI, MPI_COMM, MPI_RANK, MPI_SIZE = None, None, None, None
else:
    MPI, MPI_COMM, MPI_RANK, MPI_SIZE = None, None, None, None

if TYPE_CHECKING:
    from one_pass.opa import Opa


DEFAULT_N_DIGESTS = 40000
DEFAULT_INIT_SIZE = 1000
DEFAULT_WEIGHT = 1

METHODS = [
    "serial",
    "mp",
    "mpi",
    "tda",
]

NOW = datetime.now()

def _get_parser() -> ArgumentParser:
    parser = ArgumentParser(
        description="Program to perform TDigest performance.",
        formatter_class=RawDescriptionHelpFormatter,
    )
    parser.add_argument("method", choices=METHODS, action="store")
    parser.add_argument("-k", "--workers", type=int, action="store", default=1)
    parser.add_argument(
        "-n", "--ndigests", type=int, action="store", default=DEFAULT_N_DIGESTS
    )
    parser.add_argument(
        "-i", "--init-size", type=int, action="store", default=DEFAULT_INIT_SIZE
    )
    parser.add_argument(
        "-w", "--weight", type=int, action="store", default=DEFAULT_WEIGHT
    )
    parser.add_argument(
        "--io", action="store_true", help="Skip writing step in the benchmarking."
    )
    parser.add_argument(
        "--no-write", action="store_true", help="Perform operations, do not write benchmarking result."
    )
    return parser


def benchmark(fun) -> T.Callable[T.Any, T.Tuple[T.Any, float]]:
    def wrapper(*args, **kwargs) -> T.Tuple[T.Any, float]:
        t0 = time.time()
        res = fun(*args, **kwargs)
        t1 = time.time()
        return res, t1 - t0

    return wrapper


def init_digests(n: int, init_size: int = 0):
    digest_list = np.array([None] * n, dtype=object)

    if MPI is None or MPI_SIZE == 1:
        for j in range(n):
            digest_list[j] = TDigest(compression=1)
            if init_size:
                digest_list[j].update(np.random.normal(size=init_size))
    else:
        _mpi_init_digests(n, init_size, digest_list)

    return digest_list


def _mpi_init_digests(size: int, init_size: int, digest_list: T.Sequence[object]):
    chunk_size = size // MPI_SIZE

    # Scatter the work among all processes
    start_index = MPI_RANK * chunk_size
    if MPI_RANK == MPI_SIZE - 1:  # Last process handles the remainder
        end_index = size
    else:
        end_index = (MPI_RANK + 1) * chunk_size

    for j in range(start_index, end_index):
        digest_list[j] = TDigest(compression=1)
        if init_size:
            digest_list[j].update(np.random.normal(size=init_size))


def _get_input_and_output_structure(digests: T.Sequence[TDigest], size: int, weight: int):
    if weight == 1:
        values = np.random.normal(size=size)
        # Prepare input for parallel processing
        inputs = [(j, values[j], digests[j]) for j in range(size)]
        outputs = [None] * size

    else:
        values = np.random.normal(size=(weight, size))
        # Prepare inputs for parallel update
        inputs = [(j, values[:, j], digests[j]) for j in range(size)]
        outputs = [None] * size

    return inputs, outputs


def _update_single_digest(j, value, digest):
    """Helper function to update a single digest."""
    digest.update(value)
    return j, digest


@benchmark
def update_mp(inputs: T.Sequence[object], outputs: T.Sequence[object], workers: int):
    """Parallel version: Updates the digest for each grid point. Uses multiprocessing."""

    with ProcessPoolExecutor(max_workers=workers) as executor:
        futures = {
            executor.submit(_update_single_digest, *args): args[0] for args in inputs
        }
        for future in as_completed(futures):
            j, updated_digest = future.result()
            outputs[j] = updated_digest

    return outputs


@benchmark
def update_mpi(inputs: T.Sequence[TDigest], outputs: T.Sequence[object]):
    """Parallel version: Updates the digest for each grid point. Uses MPI."""

    num_items = len(inputs)
    chunk_size = num_items // MPI_SIZE

    # Scatter the work among all processes
    start_index = MPI_RANK * chunk_size
    if MPI_RANK == MPI_SIZE - 1:  # Last process handles the remainder
        end_index = num_items
    else:
        end_index = (MPI_RANK + 1) * chunk_size

    local_inputs = inputs[start_index:end_index]  # Slice for local processing

    local_outputs = [None] * len(local_inputs)
    for i, args in enumerate(local_inputs):
        local_outputs[i] = _update_single_digest(*args)

    gathered_outputs = MPI_COMM.gather(local_outputs, root=0)

    # Root process has MPI_RANK 0
    if MPI_RANK == 0:
        flat_outputs = [item for sublist in gathered_outputs for item in sublist]
        for idx, updated_digest in flat_outputs:
            outputs[idx] = updated_digest

    # Other processes keep only their local outputs
    else:
        outputs[start_index:end_index] = local_outputs


@benchmark
def update_serial(inputs: T.Sequence[object], outputs: T.Sequence[object]):
    for args in inputs:
        j, value = _update_single_digest(*args)
        outputs[j] = value


@benchmark
def update_tda(tda: TDigestArray, data: NDArray):
    tda.update(data)


@benchmark
def update_write_mp(
    inputs: T.Sequence[object],
    outputs: T.Sequence[object],
    workers: int,
    output_dir: Path,
    ):
    """Parallel version: Updates the digest for each grid point. Uses multiprocessing."""

    with ProcessPoolExecutor(max_workers=workers) as executor:
        futures = {
            executor.submit(_update_single_digest, *args): args[0] for args in inputs
        }
        for future in as_completed(futures):
            j, updated_digest = future.result()
            outputs[j] = updated_digest

    with open(output_dir / "outputs_mp.pkl", "wb") as bf:
        pickle.dump(outputs, bf)

    return outputs


@benchmark
def update_write_mpi(inputs: T.Sequence[TDigest], outputs: T.Sequence[object], output_dir: Path):
    """Parallel version: Updates the digest for each grid point.
    Uses MPI. Does not gather. Writes to file.
    """

    num_items = len(inputs)
    chunk_size = num_items // MPI_SIZE

    # Scatter the work among all processes
    start_index = MPI_RANK * chunk_size
    if MPI_RANK == MPI_SIZE - 1:  # Last process handles the remainder
        end_index = num_items
    else:
        end_index = (MPI_RANK + 1) * chunk_size

    local_inputs = inputs[start_index:end_index]  # Slice for local processing

    local_outputs = [None] * len(local_inputs)
    for i, args in enumerate(local_inputs):
        local_outputs[i] = _update_single_digest(*args)

    # Do not gather outputs - simply set local outputs to general outputs array (rest are still None)
    outputs[start_index:end_index] = local_outputs

    with open(output_dir / f"outputs_mpi{MPI_RANK:07d}.pkl", "wb") as bf:
        pickle.dump(outputs, bf)


@benchmark
def update_write_serial(inputs: T.Sequence[object], outputs: T.Sequence[object], output_dir: Path):
    for args in inputs:
        j, value = _update_single_digest(*args)
        outputs[j] = value

    with open(output_dir / "outputs_serial.pkl", "wb") as bf:
        pickle.dump(outputs, bf)


@benchmark
def update_write_tda(tda: TDigestArray, data: NDArray, output_dir: Path):
    tda.update(data)

    with open(output_dir / "outputs_tda.pkl", "wb") as bf:
        pickle.dump(tda, bf)


def compute_benchmarks(
    method: str,
    ndigests: int,
    init_size: int,
    weight: int,
    workers: int,
    include_io: bool,
    output_dir: Path | None,
) -> float:

    if method == "serial":
        digests = init_digests(ndigests, init_size)
        inputs, outputs = _get_input_and_output_structure(digests, ndigests, weight)
        if include_io:
            assert output_dir is not None
            _, t = update_write_serial(inputs, outputs, output_dir)
        else:
            _, t = update_serial(inputs, outputs)

    elif method == "mp":
        digests = init_digests(ndigests, init_size)
        inputs, outputs = _get_input_and_output_structure(digests, ndigests, weight)
        if include_io:
            assert output_dir is not None
            _, t = update_write_mp(inputs, outputs, workers, output_dir)
        else:
            _, t = update_mp(inputs, outputs, workers)

    elif method == "mpi":
        if MPI is None:
            raise RuntimeError(
                "MPI wasn't initialized. Please, make sure that mpi4py is installed and "
                "that the script was launched using mpirun or mpiexec."
            )
        digests = init_digests(ndigests, init_size)
        inputs, outputs = _get_input_and_output_structure(digests, ndigests, weight)
        if include_io:
            assert output_dir is not None
            _, t = update_write_mpi(inputs, outputs, output_dir)
        else:
            _, t = update_mpi(inputs, outputs)

    elif method == "tda":
        if TDigestArray is None:
            raise ImportError(
                "Forked version of crick implementing TDigestArray should be installed."
            )
        tda = TDigestArray(ndigests)
        tda.update(np.random.normal(size=(init_size, ndigests)))
        tda_update_data = np.random.normal(size=(weight, ndigests))
        if include_io:
            assert output_dir is not None
            _, t = update_write_tda(tda, tda_update_data, output_dir)
        else:
            _, t = update_tda(tda, tda_update_data)

    else:
        raise ValueError(f"Wrong method: {method}.")

    return t


def print_stats(**kwargs):
    for key, value in kwargs.items():
        title_width = (max(len(key), 70))
        underline = "=" * title_width
        title = f"{underline}\n{key.upper():^{title_width}s}\n{underline}"
        print(title + "\n\n", end="")
        pprint(value)
        print("\n\n", end="")


def write_stats(path: Path, dry=False, **kwargs):
    if dry:
        print_stats(**kwargs)
        return
    try:
        path.mkdir(parents=True, exist_ok=False)
    except FileExistsError as exc:
        raise FileExistsError(
            "Benchmark directory already exists!"
        ) from exc

    for key, value in kwargs.items():
        with open(path / f"{key}.yaml", "w", encoding="utf-8") as wf:
            yaml.safe_dump(value, wf)

def main():

    args = _get_parser().parse_args()

    try:
        benchmark_dir = Path(os.environ["BENCHMARK_DIR"]).resolve()
        benchmark_dir = benchmark_dir / "crick" / NOW.strftime(r"%Y%m%d_%H%M%S_%f")
    except KeyError as exc:
        benchmark_dir = None
        if not args.no_write:
            raise EnvironmentError(
                "Please set environment variable 'BENCHMARK_DIR'."
            ) from exc

    try:
        output_dir = Path(os.environ["OUTPUT_DIR"]).resolve()
        if (not output_dir.exists()) and args.io:
            raise FileNotFoundError(
                f"Please create output_directory: {output_dir}"
            )
        output_dir = output_dir / "crick" / NOW.strftime(r"%Y%m%d_%H%M%S_%f")
        output_dir.mkdir(exist_ok=False, parents=True)
    except KeyError as exc:
        output_dir = None
        if args.io:
            raise EnvironmentError(
                "Please set environment variable 'OUTPUT_DIR'."
            ) from exc

    # Main function to compute the benchmark
    benchmark_time = compute_benchmarks(
        method=args.method,
        ndigests=args.ndigests,
        init_size=args.init_size,
        weight=args.weight,
        workers=args.workers,
        include_io=args.io,
        output_dir=output_dir,
    )

    # Write results. For MPI, only root process writes
    if MPI is not None and MPI_SIZE > 1 and MPI_RANK > 0:
        # MPI installed, MPI enabled, non-root process.
        sys.exit(0)

    if MPI is not None and MPI_SIZE > 1:
        workers = MPI_SIZE
    else:
        workers = args.workers

    write_stats(
        path=benchmark_dir,
        dry=args.no_write,
        datetime=NOW.isoformat(" "),
        args={
            "method": args.method,
            "ndigests": args.ndigests,
            "init_size": args.init_size,
            "weight": args.weight,
            "workers": workers,
            "include_io": args.io,
        },
        argv=list(sys.argv),
        environ=dict(os.environ),
        benchmark=benchmark_time,
    )


if __name__ == "__main__":
    main()
