"""Adapted from
https://earth.bsc.es/gitlab/digital-twins/de_340-2/workflow/-/blob/main/lib/runscript/run_opa.py
"""

from argparse import ArgumentParser
import cProfile
import os
import sys

from pprint import pprint

from one_pass.opa import Opa
from opa_tools import get_request, benchmark_setup, benchmark_add
from gsv import GSVRetriever
import dask.distributed

try:
    from mpi4py import MPI
    MPI_COMM = MPI.COMM_WORLD
    MPI_RANK = MPI_COMM.Get_rank()
    MPI_SIZE = MPI_COMM.Get_size()
except ImportError:
    MPI, MPI_COMM, MPI_RANK, MPI_SIZE = None, None, None, None

data = None
opa_stat = None
request = None
gsvrequest = None
oparequest = None
parrequest = None


def _get_parser():
    parser = ArgumentParser(description="Runscript for OPA job.")
    parser.add_argument(
        "-r", "--request", type=str, required=True, help="Input request number"
    )
    parser.add_argument(
        "--read_from_databridge", type=bool, help="Read from databridge", default=False
    )
    parser.add_argument(
        "-b",
        "--benchmark",
        action="store_true",
        help="Switch on benchmarking",
    )
    parser.add_argument(
        "-d",
        "--dummy",
        action="count",
        help=(
            "Stop before running. Useful for testing. "
            "-d prevents computation, "
            "-dd prevents data retrieval."
        ),
        default=0,
    )
    return parser


def run_bias_adjustment(oparequest):  # type: ignore
    raise NotImplementedError(
        "OPA with bias adjustment not implemented in OPA-Tools package. "
        "Refer to the original script for implementation."
    )


def run(
    gsvrequest, oparequest, parrequest, *, benchmark, read_from_databridge, dummy=False
):
    global data, opa_stat

    profile = None
    # TODO: Call always benchmark_setup, with "benchmark" argument, which
    # returns profile, benchmark_path and client. They all support the same
    # methods, but do nothing and return None. This way, we avoid the verbose
    # checking "if benchmark".
    if benchmark:
        profile = cProfile.Profile()
        benchmark_path = benchmark_setup(
            benchmark=benchmark,
            environ=True,
            gsvrequest=gsvrequest,
            oparequest=oparequest,
            parrequest=parrequest,
        )
    # TODO: MAKE DEPENDANT ON A KEY IN PARREQUEST

    gsv = GSVRetriever()
    data = gsv.request_data(gsvrequest, use_stream_iterator=read_from_databridge)

    if data is None:
        raise SystemExit("No data was retreived with passed request pair.")

    if parrequest.get("forcemem", False):
        data.load()

    if dummy > 0:
        raise KeyboardInterrupt(f"Interrupted by flag dummy={dummy}")

    if benchmark:
        profile.enable()

    opa_stat = Opa(oparequest, workers=parrequest.get("workers", 1))
    opa_stat.compute(data)

    if benchmark:
        profile.disable()
        dask_client = dask.distributed.Client()
        profile.dump_stats(benchmark_path / "opa.prof")  # type: ignore
        benchmark_add(
            benchmark_path,
            dask_worker_logs=dask_client.get_worker_logs(),
            dask_stream=dask_client.get_task_stream(),
            dask_scheduler_info=dask_client.scheduler_info(),
            dask_config=dask.config.config,
        )


# Assuming GSVRetriever, Opa, benchmark_setup, benchmark_add, etc. are already imported.


def run_mpi(
    gsvrequest, oparequest, parrequest, *, benchmark, read_from_databridge, dummy=False
):

    global data, opa_stat

    if MPI is None:
        raise ImportError("mpi4py not installed.")

    profile = None
    benchmark_path = None

    # Setup benchmark (only on MPI_RANK 0, shared info)
    if benchmark and MPI_RANK == 0:
        profile = cProfile.Profile()
        benchmark_path = benchmark_setup(
            benchmark=benchmark,
            environ=True,
            gsvrequest=gsvrequest,
            oparequest=oparequest,
            parrequest=parrequest,
            MPI_size=MPI_SIZE,
        )

    # Only MPI_RANK 0 retrieves data
    if MPI_RANK == 0:
        gsv = GSVRetriever()
        data = gsv.request_data(gsvrequest, use_stream_iterator=read_from_databridge)
        if data is None:
            raise SystemExit("No data was retrieved with passed request pair.")
        if parrequest.get("forcemem", False):
            data.load()
    else:
        data = None

    # Broadcast data to all ranks (assuming it's picklable)
    data = MPI_COMM.bcast(data, root=0)

    if dummy > 0:
        raise KeyboardInterrupt(f"Interrupted by flag dummy={dummy}")

    if benchmark and MPI_RANK == 0:
        profile.enable()

    # Compute (can be parallel if Opa supports internal parallelization)
    opa_stat = Opa(oparequest, workers=parrequest.get("workers", 1))
    opa_stat.compute(data)

    if benchmark and MPI_RANK == 0:
        profile.disable()
        dask_client = dask.distributed.Client()
        profile.dump_stats(benchmark_path / "opa.prof")  # type: ignore
        benchmark_add(
            benchmark_path,
            dask_worker_logs=dask_client.get_worker_logs(),
            dask_stream=dask_client.get_task_stream(),
            dask_scheduler_info=dask_client.scheduler_info(),
            dask_config=dask.config.config,
        )


def _parse_subrequest_range(s: str):
    if s.count("-") == 1:
        raise ValueError(
            f"Subrequest range should countain one symbol '-', but was '{s}'."
        )
    start_str, end_str = s.split("-")
    if not (start_str.isdigit() and end_str.isdigit()):
        raise ValueError(
            f"Subrequest range should be formed by digits, but it was '{s}'."
        )
    try:
        return int(start_str), int(end_str)
    except TypeError as exc:
        raise TypeError(
            f"Subrequest range should be convertable to ints, but was '{s}'."
        ) from exc


def main(argv=None):
    global request, gsvrequest, oparequest, parrequest

    if argv is None:
        argv = sys.argv

    args = _get_parser().parse_args(argv[1:])

    request = args.request

    gsvrequest, oparequest, parrequest = get_request(request)
    print("\nGSV REQUEST:\n============\n")
    pprint(gsvrequest)

    print("\nOPA REQUEST:\n============\n")
    pprint(oparequest)

    print("\nOPA REQUEST:\n============\n")
    pprint(parrequest)

    if MPI is not None and MPI_SIZE > 1:
        func = run_mpi
    else:
        func = run
    func(
        gsvrequest,
        oparequest,
        parrequest=parrequest,
        benchmark=(args.benchmark and args.dummy == 0),
        read_from_databridge=args.read_from_databridge,
        dummy=bool(args.dummy),
    )


if __name__ == "__main__":
    if os.environ.get("FDB_HOME") is None:
        raise EnvironmentError("Please, set variable 'FDB_HOME'.")

    main()
