from dataclasses import dataclass, field
from functools import lru_cache
import yaml
import subprocess
import typing as T
from io import StringIO
from tqdm import tqdm

import multiprocessing as mp

FILE = ".gitlab-ci.yml"

@lru_cache
def load_yaml():
    with open(FILE, "r", encoding="utf-8") as f:
        return yaml.safe_load(f)

@dataclass
class ProcessResult:
    out: str = ""
    err: str = ""
    retcodes: T.Sequence[int] = field(default_factory=list)
    failed: bool = False


def get_scripts_by_stage(stage: str):
    config = load_yaml()
    return {
        job: data["script"]
        for job, data in config.items()
        if isinstance(data, dict) and data.get("stage") == stage
    }

def process(scripts):
    result = ProcessResult()

    completed = []
    for script in scripts:
        # Split the script into command and arguments
        args = script.split()

        # Replace "venv/" with ".venv/" in each argument
        filtered_args = [arg.replace("venv/", ".venv/") for arg in args]

        # If "source" is part of the command, run it within the same shell using bash -c
        if "source" in filtered_args:
            # We need to run 'source' within bash -c to ensure the environment is modified
            subprocess.run(f"bash -c 'source {' '.join(filtered_args[1:])}'", shell=True, check=True)
        else:
            # Run the command normally (without 'source')
            completed = subprocess.run(filtered_args, check=False, capture_output=True)
            result.out += completed.stdout.decode()
            result.err += completed.stderr.decode()
            result.retcodes += [completed.returncode]
            result.failed = result.failed or (completed.returncode != 0)

    return result


def main():
    lint_jobs = get_scripts_by_stage("lint")
    
    with mp.Pool(4) as pool:
        results = list(
            tqdm(
                pool.imap(process, lint_jobs.values()),
                total=len(lint_jobs)
            )
        )
    
    with (
        open("lintreport.out", "w", encoding="utf-8") as outf,
        open("lintreport.err", "w", encoding="utf-8") as errf
    ):
        for (job, scripts), result in zip(lint_jobs.items(), results):
            title = job
            if result.failed:
                title = "[ ERROR ] " + title
            title += "\n" + "=" * len(title)
            script_header = "\n".join(f"\t$ {script}" for script in scripts)
            outf.write(title + "\n\n")
            errf.write(title + "\n\n")
            outf.write(script_header + "\n\n")
            errf.write(script_header + "\n\n")
            outf.write(result.out)
            errf.write(result.err)
            outf.write("\n" * 3)
            errf.write("\n" * 3)
    

if __name__ == "__main__":
    main()
